@@ -504,6 +504,18 @@ gkr_pk_privkey_get_attribute (GkrPkObject* obj, CK_ATTRIBUTE_PTR attr)
 	return GKR_PK_OBJECT_CLASS (gkr_pk_privkey_parent_class)->get_attribute (obj, attr);
 }
 
+static gboolean
+gkr_pk_privkey_import (GkrPkObject *obj)
+{
+	GkrPkPrivkey *key = GKR_PK_PRIVKEY (obj);
+	GkrPkPubkey *pub;
+	
+	/* This stores all necessary information in the indexes */
+	pub = get_public_key (key, TRUE);
+	
+	return GKR_IS_PK_PUBKEY (pub);
+}
+
 static guchar*
 gkr_pk_privkey_serialize (GkrPkObject *obj, const gchar *password, gsize *n_data)
 {
@@ -518,6 +530,20 @@ gkr_pk_privkey_serialize (GkrPkObject *obj, const gchar *password, gsize *n_data
 	return gkr_pkix_serialize_private_key_pkcs8 (key->priv->s_key, password, n_data);
 }
 
+static void
+gkr_pk_privkey_lock (GkrPkObject *obj)
+{
+	GkrPkPrivkey *key = GKR_PK_PRIVKEY (obj);
+
+	if (!key->priv->s_key)
+		return;
+	
+	gcry_sexp_release (key->priv->s_key);
+	key->priv->s_key = NULL;
+	
+	initialize_from_key (key);
+}
+
 static void
 gkr_pk_privkey_dispose (GObject *obj)
 {
@@ -536,13 +562,13 @@ gkr_pk_privkey_finalize (GObject *obj)
 {
 	GkrPkPrivkey *key = GKR_PK_PRIVKEY (obj);
 
+	g_assert (!key->priv->pubkey);
+
 	gcry_sexp_release (key->priv->s_key);
 	key->priv->s_key = NULL;
-	
-	initialize_from_key (key);
-	
-	g_assert (!key->priv->pubkey);
-	g_assert (!key->priv->numbers);
+
+	gcry_sexp_release (key->priv->numbers);
+	key->priv->numbers = NULL;	
 	
 	G_OBJECT_CLASS (gkr_pk_privkey_parent_class)->finalize (obj);
 }
@@ -557,7 +583,9 @@ gkr_pk_privkey_class_init (GkrPkPrivkeyClass *klass)
 	
 	parent_class = GKR_PK_OBJECT_CLASS (klass);
 	parent_class->get_attribute = gkr_pk_privkey_get_attribute;
+	parent_class->import = gkr_pk_privkey_import;
 	parent_class->serialize = gkr_pk_privkey_serialize;
+	parent_class->lock = gkr_pk_privkey_lock;
 	
 	gobject_class = (GObjectClass*)klass;
 	gobject_class->get_property = gkr_pk_privkey_get_property;
